/*
    Copyright (C) 2004 Mark Boyd

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include<stdio.h>
#include<string.h>
#include<stdlib.h>


#include<algorithm>

#include "uiobject.h"

using std::min;
using std::max;

BITMAP *uiobjectlist::vscreen;
uiobjectlist uiobject::uol;

uiobjectlist::uiobjectlist()
{
int a;
start=active=0;
vscreen=0;
nobjects=0;
waitfun=&rest50;
for(a=0;a<256;a++)
  catchchar[a]=catchechar[a]=&donothing;
}

void uiobjectlist::mvs()
{
if(vscreen)destroy_bitmap(vscreen);
vscreen=create_bitmap(SCREEN_W,SCREEN_H);
}

void uiobjectlist::kvs()
{
if(vscreen)
  {
  destroy_bitmap(vscreen);
  vscreen=0;
  }
}

void uiobjectlist::add(uiobject *u)
{
u->next=start;
start=u;
nobjects++;
}

void uiobjectlist::remove(uiobject *u)
{
uiobject *scan;
scan=start;
if(start==u)start=start->next;
else
  {
  while(scan->next!=u)scan=scan->next;
  scan->next=u->next;
  }
if(active==u)active=0;
nobjects--;
}

void uiobjectlist::clickfun(int x,int y)
{
uiobject *scan;
scan=start;
while(scan)
  {
  scan->clickfun(x,y);
  if(scan->isactive)break;
  scan=scan->next;
  }
}

uiobject *uiobjectlist::getelt(int n)
{
int a;
uiobject *scan;
scan=start;
for(a=n+1;a<nobjects;a++)scan=scan->next;
return scan;
}


void uiobjectlist::wait(void)
{

show_mouse(screen);
while(keypressed())readkey();
stopwaiting=0;
while(!stopwaiting)
  {
  while(mouse_b)(*waitfun)();
  while(!(mouse_b||keypressed()))(*waitfun)();
  show_mouse(0);
  if(keypressed())
    {
    int k=readkey();
    k&=0xffff;
    if(!(k%256))
      {
      (*catchechar[k>>8])();
      if(active)active->echarfun(k>>8);
      }
    else
      {
      (*catchchar[k%256])();
      if(active)active->charfun(k%256);
      };
    }
  else clickfun(mouse_x,mouse_y);
  show_mouse(screen);
  }
show_mouse(0);
}

void uiobjectlist::rest50(void){rest(50);}

void uiobjectlist::donothing(void){}

uiobject::uiobject(int l,int t,int r,int b,int c)
{
left=l;top=t;right=r;bottom=b;colour=c;
parent=0;
nkids=0;
kids=0;
isactive=0;
uol.add(this);
}

uiobject::~uiobject()
{
while(kids)delete(kids[0]);
if(parent) parent->unspawn(this);
uol.remove(this);
}

void uiobject::activate(void)
{
if(uol.active)uol.active->deactivate();
isactive=1;
uol.active=this;
redraw();
}

void uiobject::deactivate(void)
{
isactive=0;
if (uol.active==this) uol.active=0;
redraw();
}

void uiobject::vclickfun(int,int)
{
while(mouse_b)rest(50);
}

void uiobject::clickfun(int x,int y)
{
if(x>=left&&x<=right&&y>=top&&y<=bottom)
  {
  activate();
  vclickfun(x,y);
  }
else if(isactive)deactivate();
}


void uiobject::spawn(uiobject *u)
{
int a;
uiobject * *t;
t=new uiobject *[nkids+1];
for(a=0;a<nkids;a++) t[a]=kids[a];
t[nkids]=u;
if(kids)delete[]kids;
kids=t;
nkids++;
u->parent=this;
}

void uiobject::unspawn(uiobject *u)
{
int a,b;
uiobject * *t;
if (nkids==1)
  {
  delete[]kids;
  nkids=0;
  kids=0;
  }
else
  {
  t=new uiobject *[nkids-1];
  for(a=b=0;a<nkids;a++)
    {
    if(kids[a]!=u)
      {
      t[b]=kids[a];
      b++;
      }
    }
  delete[]kids;
  kids=t;
  nkids--;
  }
u->parent=0;//superfluous?
}


void uiobject::redraw(void)
{
int a;
vredraw();
for(a=0;a<nkids;a++)kids[a]->redraw();
}

void uiobject::mover(int dx,int dy)
{
int a;
dx=max(-left,min(dx,SCREEN_W-1-right));
dy=max(-top,min(dy,SCREEN_H-1-bottom));
left+=dx;right+=dx;
top+=dy;bottom+=dy;
for(a=0;a<nkids;a++)kids[a]->mover(dx,dy);
}

int uiobject::getl(void){return left;}
int uiobject::gett(void){return top;}

boringbox::boringbox(int l,int t,int r,int b,int c,const char *ti)
:uiobject(l,t,r,b,c)
{
title=new char[strlen(ti)+1];
strcpy(title,ti);
blit(screen,uol.vscreen,0,0,0,0,SCREEN_W,SCREEN_H);
}

boringbox::~boringbox()
{
vsync();
blit(uol.vscreen,screen,0,0,0,0,SCREEN_W,SCREEN_H);
delete[]title;
}

void boringbox::vclickfun(int x,int y)
{
show_mouse(screen);
if(y<=top+11)
  {
  while(mouse_b)
    {
    rest(20);
    if(mouse_x!=x||mouse_y!=y)
      {
      show_mouse(0);
      vsync();
      blit(uol.vscreen,screen,0,0,0,0,SCREEN_W,SCREEN_H);
      mover(mouse_x-x,mouse_y-y);
      redraw();
      show_mouse(screen);
      x=mouse_x;y=mouse_y;
      }
    }
  }
show_mouse(0);
}

void boringbox::vredraw()
{
rect(screen,left,top,right,bottom,colour);
rectfill(screen,left+1,top+1,right-1,bottom-1,0);
hline(screen,left+1,top+11,right-1,colour);
textout(screen,font,title,(left+right)/2-4*strlen(title),top+2,colour);
}

textbox::textbox(int l,int t,int n,int c,int ic,const char *i)
:uiobject(l,t,l+2+8*n,t+11,c)
{
icolour=ic;
the_string=new char[n+1];
length=n;
strcpy(the_string,i);
cursorpos=strlen(the_string);
}

textbox::~textbox()
{
delete[]the_string;
}

char *textbox::getstr(void){return the_string;}

void textbox::vclickfun(int x,int)
{
cursorpos=min((x-left+2)/8,int(strlen(the_string)));
redraw();
}

void textbox::charfun(char c)
{
char *cp;
switch(c)
  {
  case KEY_ENTER:
  break;
  case 8: //delete
  if(cursorpos)
    {
    cursorpos--;
    cp=the_string+cursorpos;
    while((*cp++=*(cp+1)));
    }
  break;
  default:
  if ((strlen(the_string))<length)
    {
    cp=the_string+strlen(the_string)+1;
    while(cp-the_string>(int)cursorpos)*cp--=*(cp-1);
    the_string[cursorpos]=c;
    cursorpos++;
    }
  }
redraw();
}

void textbox::echarfun(char c)
{
char *cp;
switch(c)
  {
  case KEY_LEFT:
  if(cursorpos)cursorpos--;
  break;
  case KEY_RIGHT:
  if(cursorpos<strlen(the_string))cursorpos++;
  break;
  case KEY_DEL:
  if(strlen(the_string)>cursorpos)
    {
    cp=the_string+cursorpos;
    while((*cp++=*(cp+1)));
    }
  default:
  break;
  }
redraw();
}

void textbox::vredraw(void)
{
rect(screen,left,top,right,bottom,isactive?colour:icolour);
rectfill(screen,left+1,top+1,right-1,bottom-1,0);
textout(screen,font,the_string,left+2,top+2,isactive?colour:icolour);
if(isactive)vline(screen,left+1+8*cursorpos,top+1,bottom-1,icolour);
}

text::text(int l,int t,int col,const char *te)
:uiobject(l,t,l+6*strlen(te),t+8,col)
{
the_text=new char[strlen(te)+1];
strcpy(the_text,te);
}

text::~text()
{
delete[]the_text;
}

void text::vredraw(void)
{
textout(screen,font,the_text,left,top,colour);
}

button::button(int l,int t,int r,int b,int co,char *c,void (*d)(button *))
:uiobject(l,t,r,b,co)
{
strin=c;
doit=d;
}

button::~button(){}

void button::vclickfun(int,int)
{
doit(this);
while(mouse_b)rest(50);
}

void button::vredraw(void)
{
rect(screen,left,top,right,bottom,colour);
textout(screen,font,strin,(left+right)/2-4*strlen(strin)+1,(top+bottom)/2-3,colour);
}

menu::menu(int l,int t,int co,int no,int cin,const char *const *stuff)
:uiobject(l,t,l+8*maxlen(no,stuff)+3,t+11,co)
{
int a;
noptions=no;
current=cin;
text=new char *[no];
for(a=0;a<no;a++)
  {
  text[a]=new char[strlen(stuff[a])+1];
  strcpy(text[a],stuff[a]);
  }
}

menu::~menu()
{
int a;
for(a=0;a<noptions;a++) delete[]text[a];
delete[]text;
}

int menu::getcurrent(void)
{
return current;
}

void menu::vclickfun(int,int)
{
int a,f;
f=min(0,SCREEN_H-1-top-11*noptions);
store=create_bitmap(right-left+1,11*noptions+1);
blit(screen,store,left,top+f,0,0,right-left+1,11*noptions+1);
for(a=0;a<noptions;a++)
  {
  rect(screen,left,top+11*a+f,right,top+11*(a+1)+f,colour);
  rectfill(screen,left+1,top+11*a+1+f,right-1,top+11*(a+1)-1+f,0);
  textout(screen,font,text[a],left+2,top+11*a+2+f,colour);
  }
show_mouse(screen);
while(mouse_b){rest(50);}
current=max(0,min((mouse_y-top-f)/11,noptions-1));
show_mouse(0);
blit(store,screen,0,0,left,top+f,right-left+1,11*noptions+1);
redraw();
}

void menu::vredraw(void)
{
rect(screen,left,top,right,bottom,colour);
rectfill(screen,left+1,top+1,right-1,bottom-1,0);
textout(screen,font,text[current],left+2,top+2,colour);
}


hslider::hslider(int l,int t,int r,int b,int s,int c1,int c2)
:uiobject(l,t,r,b,c1)
{
left=l;top=t;bottom=b;right=r;
maximum=right-left-1;
state=s;
col2=c2;
}

int hslider::getstate(void){return state;}

void hslider::vclickfun(int,int)
{
show_mouse(screen);
while(mouse_b)
  {
  uiobject::uol.waitfun();
  state=min(maximum,max(0,mouse_x-left-1));
  show_mouse(0);
  if(state)rectfill(screen,left+1,top+1,left+state,bottom-1,col2);
  if(state!=maximum)rectfill(screen,left+state+1,top+1,right-1,bottom-1,0);
  show_mouse(screen);
  }
show_mouse(0);
}

void hslider::vredraw(void)
{
rect(screen,left,top,right,bottom,colour);
if(state)rectfill(screen,left+1,top+1,left+state,bottom-1,col2);
if(state!=maximum)rectfill(screen,left+state+1,top+1,right-1,bottom-1,0);
}

int maxlen(int n,const char *const *s)
{
int a,out;
out=0;
for(a=0;a<n;a++)out=max(out,(int)strlen(s[a]));
return out;
}




